from os import unlink


def delete_file(paths, pattern, parse_path, name, force=False):
    deleted = False
    for path in paths:
        nscl, file = parse_path(path)
        if pattern.test(nscl, quiet=True):
            if not deleted:
                print('DELETIONS:')
                deleted = True
            print('  ', file)
            if force:
                unlink(path)
    if not deleted:
        print(f'UNCHANGED {name}')



import re
from glob import glob
from os.path import join, basename

from ewconfig.prune.file import delete_file

NSCL = re.compile(r'([A-Z0-9]+\.[A-Z0-9]+\.[A-Z0-9]+\.[A-Z0-9]+)\.sac')


def parse_path(path):
    file = basename(path)
    nscl = NSCL.match(file).group(1)
    return nscl, file


def delete_response(dir, pattern, force=False):
    delete_file(glob(join(dir, 'eqk', 'response', '*.sac')), pattern, parse_path, 'response', force=force)



import re
from glob import glob
from os.path import join

from ewconfig.prune.file import delete_file

FILE = re.compile(r'.*/(SAC_PZs_.*)')
NSCL = re.compile(r'SAC_PZs_([A-Z0-9]+_[A-Z0-9]+_[A-Z0-9]+_[A-Z0-9]*)_\d+\.')


def parse_path(path):
    file = FILE.match(path).group(1)
    nscl = NSCL.match(file).group(1).replace('_', '.')
    return nscl, file


def delete_sacp(dir, pattern, force=False):
    delete_file(glob(join(dir, 'SAC_PZs_*')), pattern, parse_path, 'SACP', force=force)



from glob import glob
from logging import getLogger
from os.path import join, basename, sep
from re import compile, escape
from shutil import copy

from .merge import merge
from ewconfig.lib.file import assert_dir, assert_equal_contents, MissingDir

log = getLogger(__name__)


def merge_sac(old, new, merged, prefer, source):
    log.debug('Starting merge of SAC data')
    merge(old, new, merged, prefer, parse_sac, assert_equal_contents, write_sac, source)
    log.debug('Ending merge of SAC data')


def parse_sac(dir):
    return {nscl: path for nscl, path in glob_sac(dir)}


def glob_sac(dir):
    if dir:
        try:
            eqk_response = assert_dir(join(dir, 'eqk', 'response'))
            for file in glob(join(eqk_response, '*.sac')):
                match = NSCL_sac.match(file)
                if match:
                    yield match.group(1), file
                else:
                    log.warning(f'Could not parse {file}')
        except MissingDir:
            log.warning(f'No eqk/response in {dir}')


def write_sac(merged, nscls_map):
    dest_eqk_response = assert_dir(join(merged, 'eqk', 'response'), create=True)
    for nscl, path in nscls_map.items():
        log.debug(f'Copying {path} for {nscl} to {dest_eqk_response}')
        copy(path, join(dest_eqk_response, basename(path)))


NSCL_sac = compile(r'^.*' + escape(sep) + r'([A-Z0-9]+\.[A-Z0-9]+\.[A-Z0-9]+\.[-A-Z0-9]+)\.sac$')


from logging import getLogger
from os.path import exists, basename
from shutil import copytree, copy2

log = getLogger(__name__)


def copy_rest(old, merged):
    if old:
        log.info(f'Copying rest of {old} to {merged}')
        copytree(old, merged, ignore=_copy_filter(old), dirs_exist_ok=True, copy_function=_error_on_overwrite)
    else:
        log.warning('No additional config (specify --old)')


def _error_on_overwrite(src, dst):
    """
    This is to highlight any failure in assumptions made.  Issue 28 occurred because we were
    over-writing files.  The bug was elsewhere (in _copy_filter below), but we would have
    detected it here.
    """
    if exists(dst):
        raise Exception(f'Refusing to copy {src} to {dst} because file already exists')
    else:
        copy2(src, dst)


def _copy_filter(old):

    def select(files, *exclude):
        return [file for file in files if file in exclude]

    def filter(dir, files):
        if dir == old:
            return select(files, 'chan', 'stationxml')
        elif basename(dir) == 'eqk':
            return select(files, 'response')
        else:
            return []

    return filter



from abc import ABC, abstractmethod
from fnmatch import translate
from logging import getLogger
from re import compile

log = getLogger(__name__)


class BaseFilter(ABC):

    @abstractmethod
    def _default_includes(self):
        raise NotImplemented()

    def __init__(self, name, includes=None, excludes=None):
        self.__name = name
        if isinstance(includes, str) or isinstance(excludes, str):
            raise Exception('Filter takes lists of includes and excludes')
        if not includes: includes = self._default_includes()
        if not excludes: excludes = []
        self.__str = f'{name}: include ' + ','.join(includes)
        if excludes: self.__str += '; exclude ' + ','.join(includes)
        self.__includes = [compile(translate(include)) for include in includes]
        self.__excludes = [compile(translate(exclude)) for exclude in excludes]

    def __call__(self, items, key=None, quiet=False):
        logged = set()
        for item in items:
            if self.test(item, key=key, quiet=quiet, logged=logged):
                yield item

    def test(self, item, key=None, quiet=False, logged=None):
        if not key: key = lambda x: x
        if logged is None: logged = set()
        value = key(item)
        if any(include.match(value) for include in self.__includes):
            if not any(exclude.match(value) for exclude in self.__excludes):
                return True
            else:
                if value not in logged:
                    if not quiet: log.info(f'{self.__name} {value} was explicitly excluded')
                    logged.add(value)
        else:
            if value not in logged:
                if not quiet: log.info(f'{self.__name} {value} was not included')
                logged.add(value)
        return False

    def __str__(self):
        return self.__str


class Filter(BaseFilter):
    """
    A filter for <name> that includes patterns in <includes>
    and then excludes patterns in <excludes>.
    If <includes> is missing then all are initially included.
    If <excludes> is missing then none are excluded.
    Patterns are 'globs' as supported by fnmatch module.
    """

    def _default_includes(self):
        return ['*']


class DottedFilter(BaseFilter):
    """
    Create a filter for <name> that includes patterns in <includes>
    and then excludes patterns in <excludes>.
    If <includes> is missing then none are initially included (differs from Filter)
    If <excludes> is missing then none are excluded.
    Patterns are 'globs' as supported by fnmatch module, but unlike Filter they can contain
    multiple components separated by dots.  When matching (possibly right-incomplete) NSCLs
    the number of components must match.  (In practice these are just characters!  No special
    processing is necessary!)
    """

    def _check_dots(self, pattern):
        # this shouldn't happen when called from the command line because it's verified by the type
        n = pattern.count('.')
        if n not in (1, 3):
            log.warning(f'The pattern {pattern} does not contain 2 or 4 dotted names (N.S or N.S.C.L) so will not be used')

    def __init__(self, name, includes=None, excludes=None):
        if includes:
            for include in includes:
                self._check_dots(include)
        if excludes:
            for exclude in excludes:
                self._check_dots(exclude)
        super().__init__(name, includes=includes, excludes=excludes)

    def _default_includes(self):
        return []



from collections import defaultdict
from logging import getLogger
from os.path import join
from re import match

from ewconfig.lib.file import assert_empty_dir
import ewconfig.merge.binder as binder

log = getLogger(__name__)

DEFAULT_PICK_FP_STA = '-1 -1 10.0 10.0 -1'

INST = 'INST_WILDCARD'
MOD = 'MOD_WILDCARD'
INDEX_SIZE = 10000
TANKSIZE_VAR = '${WSV_TANK_MEGS}'
TANKDIR_VAR = '${WSV_TANK_DIR}'
EWDATA_VAR = '${EW_DATA_DIR}'

NM_IN_M = 1e9


def write_pz_ew(dir, sncl, drop_comment=False, m_to_nm=1):
    with open(join(dir, "%s.%s.%s.%s.sac" % (sncl.net, sncl.sta, sncl.chan, sncl.loc)), "w") as f:
        if not drop_comment:
            f.write(sncl.keep_comment)
        f.write("ZEROS   %d\n" % len(sncl.zeros_lines))
        for line in sncl.zeros_lines:
            f.write(line)
        f.write("POLES   %d\n" % len(sncl.poles_lines))
        for line in sncl.poles_lines:
            f.write(line)
        sncl.constant = sncl.constant / m_to_nm
        if m_to_nm == 1:
            comment = 'UNITS IN M'
        elif m_to_nm == NM_IN_M:
            comment = 'UNITS IN NM'
        else:
            comment = f'UNITS UNKNOWN (FACTOR {m_to_nm:g})'
        f.write(f"* {comment}\nCONSTANT        {sncl.constant:e}\n")


def write_wsv(dir, sncls):
    with open(join(dir, 'wsv_chan_list.d'), 'w') as f:
        f.write(
            "#          names       size  (TYPE_TRACEBUF2 only)         (megabytes) (max breaks)     (full path)      Tank\n")
        for sncl in sncls:
            f.write("Tank %s %s %s %s 4096 %s %s %s %d %s/%s/%s.%s.%s.%s.wsv_tnk\n" % \
                    (sncl.sta, sncl.chan, sncl.net, sncl.loc, INST, MOD, TANKSIZE_VAR, INDEX_SIZE, EWDATA_VAR, TANKDIR_VAR, \
                     sncl.sta, sncl.chan, sncl.net, sncl.loc))


def write_pick_sta(dir, sncls, geophone=False):
    # example from Raton
    #    1     1  T25A  BHZ TA -- 3  40  3 162  500  3  .939  3.  .4  .015 5.  .9961  1200. 132.7   .8  1.5 135000. 8388608
    #    1     1  T25A  BHE TA -- 3  40  3 162  500  3  .939  3.  .4  .015 5.  .9961  1200. 132.7   .8  1.5 135000. 8388608
    #    1     1  T25A  BHN TA -- 3  40  3 162  500  3  .939  3.  .4  .015 5.  .9961  1200. 132.7   .8  1.5 135000. 8388608
    with open(join(dir, 'pick_sta.d'), 'w') as f:
        f.write(
            '#\n #                              MinBigZC       RawDataFilt    LtaFilt         DeadSta          PreEvent\n')
        f.write('# Pick  Pin    Station/   MinSmallZC   MaxMint           StaFilt       RmavFilt           AltCoda\n')
        f.write(
            '# Flag  Numb   Comp/Net   Itr1   MinPeakSize  i9  CharFuncFilt  EventThresh          CodaTerm         Erefs\n')
        f.write(
            '# ----  ----   --------   ----------------------------------------------------------------------------------\n')
        for sncl in sncls:
            if geophone or sncl.chan.startswith('E'):
                f.write("1 1 %s %s %s %s 3 40 3 300 500 0 .985 0. .0198 .002 3. .9961 1200. 100.0 .8 1.5 50000. 8388608\n" % \
                        (sncl.sta, sncl.chan, sncl.net, sncl.loc))
            else:
                f.write(
                    "1  1 %s %s %s %s 3  40  3 162  500  3  .939  3.  .4  .015 5.  .9961  1200. 132.7   .8  1.5 135000. 8388608\n" % \
                    (sncl.sta, sncl.chan, sncl.net, sncl.loc))


def expand_angle(angle, pos, neg):
    if angle < 0:
        char = neg
        angle = abs(angle)
    else:
        char = pos
    degrees = int(angle)
    minutes = (angle - degrees) * 60.
    return degrees, minutes, char


def write_hinv_sta(dir, sncls):
    # P02   XR  EHZ  37 05.0890 104 51.1720 17830.0     0.02  0.00  0.00  0.00    0.0000
    with open(join(dir, 'hinv_sta.d'), 'w') as f:
        for sncl in sncls:
            lat_deg, lat_min, lat_char = expand_angle(sncl.lat, ' ', 'S')
            lon_deg, lon_min, lon_char = expand_angle(sncl.lon, 'E', ' ')
            f.write("%-5s %2s  %3s  %2d %7.4f%1s%3d %7.4f%1s%4d0.0     0.00  0.00  0.00  0.00    0.00%2s\n" % \
                    (sncl.sta, sncl.net, sncl.chan, lat_deg, lat_min, lat_char, lon_deg, lon_min, lon_char, int(sncl.elev), sncl.loc))


def write_binder_grid(dir, sncls):
    ''' Write grid configuration for binder_ew.

    '''
    with open(join(dir, 'binder_grid.d'), 'w') as f:
        # Compute the grid boundaries.
        lat = [x.lat for x in sncls]
        lon = [x.lon for x in sncls]

        round_dec = 2
        grid_limits = binder.compute_grid_limits(lon = lon,
                                                 lat = lat,
                                                 round_dec = round_dec)

        grid_params = binder.compute_grid_parameters(lon = lon,
                                                     lat = lat)

        float_string = '{{:.{:d}f}}'.format(round_dec)
        template_lat = 'grdlat    {}    {}\n'.format(float_string,
                                                     float_string)
        template_lon = 'grdlon    {}    {}\n'.format(float_string,
                                                     float_string)
        f.write(template_lon.format(grid_limits[0],
                                    grid_limits[1]))
        f.write(template_lat.format(grid_limits[2],
                                    grid_limits[3]))
        f.write('dspace    {:f}\n'.format(grid_params['dspace']))
        f.write('rstack    {:f}\n'.format(grid_params['rstack']))
        f.write('tstack    {:f}\n'.format(grid_params['tstack']))


def write_seed_link_stas2(dir, sncls):
    ns_to_chan02 = defaultdict(set)
    for sncl in sncls:
        ns = f'{sncl.net}_{sncl.sta}'
        ns_to_chan02[ns].add(sncl.chan[0:2])
    with open(join(dir, 'slink_imports.d'), 'w') as f:
        for ns in sorted(ns_to_chan02.keys()):
            chans = " ".join(f'{chan02}?.D' for chan02 in sorted(ns_to_chan02[ns]))
            f.write(f'Stream   {ns} "{chans}"\n')


def write_pick_fp(dir, sncls, pick_fp_sta):
    #                                      threshold1
    # Pick  Pin     Sta/Comp           longTermWindow  tUpEvent
    # Flag  Numb    Net/Loc       filterWindow  threshold2
    # ----  ----    --------      -----------------------------
    #    1    00  AVG 2.4 6.0 10.0 8.0 0.2
    #
    # ----  ----   --------   ----------------------------------------------------------------------------------
    # 1  1 ACSO BH1 US 00  2.4 6.0 10.0 8.0 0.2
    with open(join(dir, 'pick_FP_sta.d'), 'w') as f:
        f.write('#                                      threshold1\n')
        f.write('# Pick  Pin     Sta/Comp           longTermWindow  tUpEvent\n')
        f.write('# Flag  Numb    Net/Loc       filterWindow  threshold2\n')
        f.write('# ----  ----    --------      -----------------------------\n')
        for sncl in sncls:
            f.write("1    1 %s %s %s %s %s\n" % \
                    (sncl.sta, sncl.chan, sncl.net, sncl.loc, pick_fp_sta))


def write_trigsta(dir, sncls):
    with open(join(dir, 'trigsta.scnl'), 'w') as f:
        for sncl in sncls:
            f.write("TrigStation %s %s %s %s\n" % (sncl.sta, sncl.chan, sncl.net, sncl.loc))


def write_tbuf2mseed(dir, sncls):
    with open(join(dir, 'tbuf2mseed.d'), 'w') as f:
        for sncl in sncls:
            f.write("Send_scnl %s %s %s %s 10\n" % (sncl.sta, sncl.chan, sncl.net, sncl.loc))


def write_carlsta(dir, sncls):
    with open(join(dir, 'carlsta.scnl'), 'w') as f:
        for sncl in sncls:
            # the 10 is a TTL.  see discussion in issue #6 - if the user wants a different
            # value they can edit the file.
            f.write("station %s %s %s %s 10\n" % (sncl.sta, sncl.chan, sncl.net, sncl.loc))


def write_fir(dir, sncls):
    # sncls are already sorted so stations are grouped together
    with open(join(dir, 'fir.scnl'), 'w') as f:
        for sncl in sncls:
            f.write("GetSCNL %s %s %s %s %s %s %s %s\n" %
                    (sncl.sta, sncl.chan, sncl.net, sncl.loc, sncl.sta, sncl.chan, sncl.net, sncl.loc))


ITYPES = {'.N': 1, '[BH]H': 2, '[SE]H': 3}


def write_stalist(dir, sncls):
    with open(join(dir, 'stalist.txt'), 'w') as f:
        for sncl in sncls:
            itype = None
            for (prefix, value) in ITYPES.items():
                if match(prefix, sncl.chan.upper()):
                    itype = value
            if itype:
                f.write("%s %s %s %s %f %f %f %f %s\n" %
                        (sncl.sta, sncl.chan, sncl.net, sncl.loc,
                         sncl.lat, sncl.lon, sncl.rate, sncl.gain, itype))
            else:
                log.warning(f'No match for {sncl} when writing stalist.txt')


def log_pz(sncl):
    log.info("%s lat=%f lon=%f elev=%f const=%e np=%d nz=%d" %
             (sncl, sncl.lat, sncl.lon,
              sncl.elev, sncl.constant, len(sncl.poles_lines), len(sncl.zeros_lines)))


def write_all(dir, sncls, drop_comment=False, m_to_nm=1, geophone=False,
              pick_fp_sta=DEFAULT_PICK_FP_STA):
    # we sort so that the output is constant even if different sources generate sncls in different orders
    sncls = sorted(sncls, key=str)
    chan = assert_empty_dir(join(dir, 'chan'), extra=' (aborting with some changes already made to the system)')
    eqk = assert_empty_dir(join(dir, 'eqk'), extra=' (aborting with some changes already made to the system)')
    write_wsv(chan, sncls)
    write_pick_sta(chan, sncls, geophone=geophone)
    write_pick_fp(chan, sncls, pick_fp_sta)
    write_hinv_sta(chan, sncls)
    write_trigsta(chan, sncls)
    write_tbuf2mseed(chan, sncls)
    write_stalist(chan, sncls)
    z_sncls = [sncl for sncl in sncls if sncl.chan.endswith('Z')]
    write_carlsta(chan, z_sncls)
    write_fir(chan, z_sncls)
    eqk_response = assert_empty_dir(join(dir, 'eqk', 'response'),
                                    extra=' (aborting with some changes already made to the system)')
    for sncl in sncls:
        write_pz_ew(eqk_response, sncl, drop_comment=drop_comment, m_to_nm=m_to_nm)
    write_seed_link_stas2(chan, sncls)

    write_binder_grid(eqk, sncls)


def add_write_args(parser, with_drop_comment=True):
    if with_drop_comment:
        parser.add_argument('-d', '--drop-comment', action="store_true",
                            help='Delete the comments from the sac output.')
    else:
        parser.set_defaults(drop_comment=False)
    parser.add_argument('-n', '--nano', action='store_const',
                        const=NM_IN_M, default=1,
                        help='Convert Constant meters into nanometers. '
                             'You must use this option if you are using the default Localmag or GMEW settings. '
                             'If you have enabled the ResponseInMeters option in localmag.d and/or gmew.d, '
                             'then you don\'t need this option.')
    parser.add_argument('-g', '--geophone', action="store_true",
                        help='Treat ALL instruments as Geophones in pick_ew output. '
                             'Otherwise broadband settings are used unless channel starts with E.')



from argparse import ArgumentParser
from logging import getLogger
from os.path import isfile, exists
from shutil import move, rmtree

from ewconfig.lib.write import DEFAULT_PICK_FP_STA
from ewconfig.merge.xml import merge_xml
from .lib.args import ParagraphHelpFormatter, add_version_args
from .lib.file import assert_dir, tmp_dir, assert_empty_dir
from .lib.filter import Filter
from .lib.log import add_log_args, make_log_from_args
from .lib.markdown import add_md_help_argument
from .merge.binder import compute_binder_grid
from .merge.chan import merge_filtered_chan, sncl_sort
from .merge.prefer import Prefer
from .merge.rest import copy_rest
from .merge.sac import merge_sac
from .stationxml2ew import add_stationxml2ew_args, main_single

log = getLogger(__name__)


def expand_xml(file, drop_comment=False, m_to_nm=1, geophone=False,
               lo_precision=False, default_response=None,
               sta_filter=Filter('Station'), chan_filter=Filter('Channel'), date=None,
               pick_fp_sta=DEFAULT_PICK_FP_STA):
    if isfile(file):
        if not file.endswith('.xml'):
            log.warning(f'Treating {file} as a station.xml file and calling stationxml2ew')
        dir = tmp_dir()
        main_single(file, dir=dir, drop_comment=drop_comment, m_to_nm=m_to_nm, geophone=geophone,
                    lo_precision=lo_precision, default_response=default_response,
                    sta_filter=sta_filter, chan_filter=chan_filter, date=date, pick_fp_sta=pick_fp_sta)
        log.debug(f'Expanded {file} into {dir}')
        return dir, file
    else:
        return file, False


def _prepare_dirs(news, old=None, merged=None, backup=None):
    if old: old = assert_dir(old, 'Old directory (--old)')
    news = [(assert_dir(new, 'New directory (--new)'), delete) for new, delete in news]
    if not merged:
        merged = tmp_dir()
        log.info(f'Merging to temporary directory {merged}')
    merged = assert_empty_dir(merged, 'Merge directory (--merged)', extra=' (aborting with no changes made)')
    mergeds = [merged]
    while len(mergeds) < len(news):
        mergeds = [tmp_dir()] + mergeds  # pad to front so final merged at end
    if backup == 'NONE':
        log.warning('No backup (--backup NONE)')
        backup = None
    if backup and exists(backup):
        raise Exception(f'--backup {backup} already exists')
    return old, news, mergeds, backup


def main(news, old=None, merged=None, backup=None, prefer=Prefer.CHECK, sort=None,
         drop_comment=False, m_to_nm=1, geophone=False, lo_precision=False, default_response=None,
         sta_filter=Filter('Station'), chan_filter=Filter('Channel'),
         date=None, comment=None, all_chan=False, pick_fp_sta=DEFAULT_PICK_FP_STA):
    accumulate = None
    try:
        if backup and not old:
            raise Exception('--backup can only be used with --old')
        if not backup and not merged:
            raise Exception('Provide --merged or --backup (the latter only with --old)')
        news = [expand_xml(new, drop_comment=drop_comment, m_to_nm=m_to_nm, geophone=geophone,
                           lo_precision=lo_precision, default_response=default_response,
                           sta_filter=sta_filter, chan_filter=chan_filter, date=date, pick_fp_sta=pick_fp_sta)
                for new in news]
        old, news, mergeds, backup = _prepare_dirs(news, old=old, merged=merged, backup=backup)
        current_old = old  # where we have currently accumulated to
        for index, ((new, xml_file), accumulate) in enumerate(zip(news, mergeds)):
            if xml_file:
                log.debug(f'Processing {new}/{xml_file}')
                source = xml_file
            else:
                log.debug(f'Processing directory {new}')
                source = new
            merge_sac(current_old, new, accumulate, prefer, source)
            merge_filtered_chan(current_old, new, accumulate, prefer, sort, source,
                                comment if index == 1 else None, all=all_chan)
            merge_xml(current_old, new, accumulate, prefer)
            if xml_file:
                log.debug(f'Deleting temporary {new} (where {xml_file} was expanded)')
                rmtree(new)
            if current_old != old:  # so a temporary intermediate
                log.debug(f'Deleting temporary {current_old} (intermediate merge)')
                rmtree(current_old)
            current_old = accumulate
        copy_rest(old, current_old)

        # Processing for which the complete inventory is needed.
        compute_binder_grid(current_old)
        
        if merged:
            return accumulate
        else:
            if backup:
                log.info(f'Moving {old} to {backup} (creating backup)')
                move(old, backup)
            else:
                log.warning(f'Deleting {old} (--backup NONE)')
                rmtree(old)
            log.info(f'Moving {accumulate} to {old} (overwriting old)')
            move(accumulate, old)
            return backup
    except:
        if accumulate:
            if exists(accumulate):
                log.warning(f'You may need to delete intermediate data in {accumulate} '
                            f'(initial data should be unchanged)')
            elif not backup:
                log.warning(f'Check {old} for corrupt data (--backup NONE)')
        raise


def main_args():
    try:
        parser = ArgumentParser(prog='ewmerge', add_help=False,
                                formatter_class=ParagraphHelpFormatter,
                                description='''A program to merge EW configurations 
(at --old and --new) selecting entries based on NSCLs.

The --new sources can be Station XML format files or EW config directories.

In the final output, files in the eqk/response and chan sub-directories are merged / filtered 
from all directories and files.  All other configuration data are taken from --old (if given).

If --merged is given, the new, merged config is placed in that directory.

If --merged is not given, but --old is used, the merged config replaces the current contents
of the --old directory.  In this case --backup (which will contain the original contents of
--old) must be given.  If you really don't want a backup use "--backup NONE".

Examples

To merge data from all the station xml files in the "input" directory, placing the EW config
in the "output" directory:

    ewmerge --new input/*.xml --merged output
    
To merge data from all the station xml files in the "input" directory into an existing "config"
directory:

    ewmerge --new input/*.xml --old config --backup old-config
    
To merge data from all the station xml files in the "input" directory with an existing "config"
directory, placing the results in a new "output" directory:

    ewmerge --new input/*.xml --old config --merged output
''')
        group = parser.add_argument_group(title='Merge parameters')
        group.add_argument('--old', help='Directory for old (full) config.')
        group.add_argument('--new', required=True, nargs='+',
                           help='Directory(s) or XML file(s) for new (chan and response) config.  '
                                'XML files are expanded by stationxml2ew (see parameters below).')
        prefer = group.add_mutually_exclusive_group()
        prefer.add_argument('--prefer-old', action='store_const', dest='prefer', const=Prefer.OLD,
                            help='If a NSCL is present in old and new, use the old entries '
                                 '(default is to require identical entries).')
        prefer.add_argument('--prefer-new', action='store_const', dest='prefer', const=Prefer.NEW,
                            help='If a NSCL is present in old and new, use the new entries.')
        group.set_defaults(prefer=Prefer.CHECK)
        group.add_argument('--merged', help='Destination directory for merged config.')
        group.add_argument('--backup', help='Backup directory for old config (if --old and no --merged; can be NONE).')
        group.add_argument('--sort', type=sncl_sort, help='Sorting order for output (eg SNCL, NS).')
        group.add_argument('--comment', nargs='?', default=None, const='',
                           help='Comment before second file (no argument for timestamp only)')
        group.add_argument('--all-chan', action='store_true',
                           help='Fail if a file in the chan directory is an unknown type (default is to ignore).')
        group.add_argument('-h', action='help', help='Show this help message and exit.')
        add_version_args(group)
        add_md_help_argument(group)
        add_log_args(group)
        group = parser.add_argument_group(title='XML expansion parameters (passed to stationxml2ew)')
        add_stationxml2ew_args(group)
        args = parser.parse_args()
        make_log_from_args(parser.prog, args)
        output = main(news=args.new, old=args.old, merged=args.merged, backup=args.backup,
                      prefer=args.prefer, sort=args.sort,
                      drop_comment=args.drop_comment, m_to_nm=args.nano, geophone=args.geophone,
                      default_response=args.default_response, date=args.date, comment=args.comment,
                      pick_fp_sta=args.pick_fp_sta, all_chan=args.all_chan,
                      sta_filter=Filter('Station', includes=args.include_sta, excludes=args.exclude_sta),
                      chan_filter=Filter('Channel', includes=args.include_chan, excludes=args.exclude_chan))
    except Exception as e:
        log.critical(e)
        exit(1)


if __name__ == '__main__':
    main_args()



from glob import glob
from logging import getLogger
from os import unlink
from os.path import join

from obspy import read_inventory

log = getLogger(__name__)


def _delete_file(file, filter, force=False):
    deleted = False

    def warn(code):
        nonlocal deleted
        if not deleted:
            print(f'DELETIONS in {file}')
            deleted = True
        print('  ' + code)

    metadata = read_inventory(file)
    for net in list(metadata):
        for sta in list(net):
            for chan in list(sta):
                nscl = f'{net.code}.{sta.code}.{chan.code}.{chan.location_code}'
                if filter.test(nscl):
                    warn(nscl)
                    if force:
                        metadata.remove(network=net.code, station=sta.code, channel=chan.code)
            ns = f'{net.code}.{sta.code}'
            if filter.test(ns):
                warn(ns)
                if force:
                    metadata.remove(network=net.code, station=sta.code)
        if force and len(net) == 0:
            log.debug(f'Deleting {net.code} as no stations remaining')
            metadata.remove(network=net.code)
    if force:
        unlink(file)
        log.info(f'Writing filtered XML to {file}')
        metadata.write(file, format='STATIONXML')


def delete_xml(dir, filter, force=False):
    for file in glob(join(dir, 'stationxml', '*', '*.xml')):
        _delete_file(file, filter, force=force)




from logging import getLogger
from os import listdir
from os.path import join, exists
from re import compile
import datetime as dt
from textwrap import wrap

from .parser import BaseParser
from .prefer import Prefer
from ewconfig.lib.file import fix_newline

log = getLogger(__name__)

PARSERS = {}


def merge_filtered_chan(old, new, merged, prefer, sort, source, comment, all=False):
    # by the time we get here we've used the filters to select NSCLs from the two directories
    all_files = set(ls_chan(old)).union(ls_chan(new))
    for file in sorted(all_files):
        if file in PARSERS:
            Parser = PARSERS[file]
            log.debug(f'Parsing {file} with {Parser}')
            new_parser = Parser(file, prefer, sort, source)
            new_parser.read(new, comment)
            if old:
                old_parser = Parser(file, prefer, sort, join(old, file))
                old_parser.read(old)
                old_parser.add(new_parser)
                old_parser.write(merged)
            else:
                new_parser.write(merged)
        else:
            if all:
                raise Exception(f'No support for {file} (see --all-chan)')
            else:
                log.warning(f'No support for {file} (ignoring; use --all-chan to fail in this case)')


def ls_chan(dir):
    if dir:
        chan = join(dir, 'chan')
        if exists(chan):
            return listdir(chan)
    return []


def sncl_sort(template):
    """
    The type used for search templates.
    """
    template = template.toupper()
    if len(set(template)) != len(template):
        raise Exception(f'Repetitions in {template}')
    for letter in template:
        if letter not in 'NSCL':
            raise Exception(f'Bad character {letter} in {template} (not S, N, C, or L)')
    return template


COMMENT = compile(r'^\s*(?:\s*|#.*)$')
END = None


class RegexParser(BaseParser):
    '''
    Files with an optional header where each line contains NSCL related data.
    Each line is matched against a pattern and compared with the pre-selected NSCLs.
    '''

    def __init__(self, line_regex, file, prefer, sort, source):
        self.__line_regex = line_regex
        self.__prefer = prefer
        self.__sort = sort
        self.__comments = {}
        self.__lines = {}
        super().__init__(file, source)

    def _match_to_key(self, match):
        raise NotImplementedError('_match_to_key')

    def _initial_comments(self, comment):
        if comment is None:
            return []
        else:
            date = dt.date.today().isoformat()
            if comment:
                comment = f'{date} {comment} Added by ewmerge, part of ewconfig.'
            else:
                comment = f'{date} Added by ewmerge, part of ewconfig.'
            return [fix_newline(line)
                    for line in wrap(comment, width=78, initial_indent='# ', subsequent_indent='# ')]

    def _read(self, file, comment=None):
        comments = self._initial_comments(comment)
        for line in file:
            line = fix_newline(line)
            if COMMENT.match(line):
                comments.append(line)
            else:
                match = self.__line_regex.match(line)
                if match:
                    key = self._match_to_key(match)
                    if key in self.__lines:
                        log.warning(f'Duplicate data in {file.name} '
                                    f'replacing [{self.__lines[key].strip()}] with [{line.strip()}]')
                    self.__lines[key] = line
                    self.__comments[key] = comments
                    comments = []
                else:
                    raise Exception(f'Could not parse "{line.strip()}" in {file.name}')
        if comments:
            self.__comments[END] = comments

    def add(self, new):
        if self._source == new._source:
            log.error('Calling twice?')
        for key, line in new.__lines.items():
            if key in self.__lines:
                if line != self.__lines[key]:
                    if self.__prefer == Prefer.OLD:
                        log.warning(f'Taking {key} from {self._source} (preferring old data over {new._source})')
                        pass  # already have the old line
                    elif self.__prefer == Prefer.NEW:
                        log.warning(f'Taking {key} from {new._source} (preferring new data over {self._source})')
                        self.__lines[key] = line
                        self.__comments[key] = new.__comments[key]
                    else:
                        raise Exception(f'Entries differ for {key} in {self._source}')
                else:
                    log.debug(f'Entries consistent for {key} in {self._source}')
            else:
                self.__lines[key] = line
                self.__comments[key] = new.__comments[key]

    def _sorted_items(self):
        # this returns (key, lines) where lines includes preceding comments
        # list() avoids errors on deletion (mutation)
        if not self.__sort:
            prev_comments = []
            for key in list(self.__lines.keys()):
                if key in self.__comments:
                    lines = list(self.__comments[key])  # copy so we don't mutate when appending line!
                    if lines == prev_comments:
                        log.debug(f'Skipping duplicate comments in {self._file}')
                        lines = []
                    elif lines:
                        prev_comments = list(lines)
                else:
                    lines = []
                lines.append(self.__lines[key])
                yield key, lines
        else:
            if self.__comments:
                log.warning(f'Dropping comments since entries are sorted ({self._source})')
            for key in list(self.__sorted_keys()):
                yield key, [self.__lines[key]]

    def __sorted_keys(self):
        for mapped, raw in sorted([(self.__map_key(key), key) for key in self.__lines.keys()],
                                  key=lambda pair: pair[0]):
            yield raw

    def __map_key(self, key):
        mapped_key = []
        components = key.split('.')
        for letter in self.__sort:
            try:
                index = self.KEY.index(letter)
                mapped_key.append(components[index])
            except ValueError:
                pass  # not in KEY
        return mapped_key

    def _write(self, file):
        for key, lines in self._sorted_items():
            for line in lines:
                file.write(line)

    def delete(self, pattern, force=False):
        for key, lines in self._sorted_items():
            if pattern.test(key, quiet=True):
                self._print_deletion(lines)
                if force:
                    del self.__lines[key]
                    del self.__comments[key]
        self._end_deletions()


class NSCL(RegexParser):
    '''
    Selection uses net, sta, chan and loc.
    '''

    KEY = 'NSCL'

    def _match_to_key(self, match):
        return f'{match.group("net")}.{match.group("sta")}.{match.group("chan")}.{match.group("loc")}'


class NSC(RegexParser):
    '''
    Selection uses net, sta and chan.
    '''

    KEY = 'NSC'

    def _match_to_key(self, match):
        return f'{match.group("net")}.{match.group("sta")}.{match.group("chan")}'


class NS(RegexParser):
    '''
    Selection uses net and sta.
    '''

    KEY = 'NS'

    def _match_to_key(self, match):
        return f'{match.group("net")}.{match.group("sta")}'


class BINDER(RegexParser):
    '''
    Select grdlat or grdlon.
    '''

    KEY = 'BINDER'

    def _match_to_key(self, match):
        coord_type = match.group("coord_type")
        return (self._source, coord_type)


SNC_prefix_trailing_L = compile(r'^\s*(?P<sta>[A-Z0-9]+)\s+(?P<net>[A-Z0-9]+)\s+(?P<chan>[A-Z0-9]+)\s+.*(?P<loc>(?:[A-Z0-9]{2}|--))$')
SCNL_after_2 = compile(r'^\s*\d+\s+\d+\s+(?P<sta>[A-Z0-9]+)\s+(?P<chan>[A-Z0-9]+)\s+(?P<net>[A-Z0-9]+)\s+(?P<loc>(?:[A-Z0-9]+|--))\s+')
NS_after_Stream =  compile(r'^\s*Stream\s*(?P<net>[A-Z0-9]+)_(?P<sta>[A-Z0-9]+)\s+')
SCNL_after_TrigStation = compile(r'^\s*TrigStation\s+(?P<sta>[A-Z0-9]+)\s+(?P<chan>[A-Z0-9]+)\s+(?P<net>[A-Z0-9]+)\s+(?P<loc>(?:[A-Z0-9]+|--))')
SCNL_after_station = compile(r'^\s*station\s+(?P<sta>[A-Z0-9]+)\s+(?P<chan>[A-Z0-9]+)\s+(?P<net>[A-Z0-9]+)\s+(?P<loc>(?:[A-Z0-9]+|--))')
SCNL_after_Send_scnl = compile(r'^\s*Send_scnl\s+(?P<sta>[A-Z0-9]+)\s+(?P<chan>[A-Z0-9]+)\s+(?P<net>[A-Z0-9]+)\s+(?P<loc>(?:[A-Z0-9]+|--))')
SCNL_after_Tank = compile(r'^\s*Tank\s+(?P<sta>[A-Z0-9]+)\s+(?P<chan>[A-Z0-9]+)\s+(?P<net>[A-Z0-9]+)\s+(?P<loc>(?:[A-Z0-9]+|--))')
SCNL_after_GetSCNL = compile(r'^\s*GetSCNL\s+(?P<sta>[A-Z0-9]+)\s+(?P<chan>[A-Z0-9]+)\s+(?P<net>[A-Z0-9]+)\s+(?P<loc>(?:[A-Z0-9]+|--))')
SNCL = compile(r'^\s*(?P<sta>[A-Z0-9]+)\s+(?P<net>[A-Z0-9]+)\s+(?P<chan>[A-Z0-9]+)\s+(?P<loc>(?:[A-Z0-9]{2}|--))')

binder_grid_coord_min_max = compile(r'^\s*(?P<coord_type>(grdlat|grdlon))\s+.*')


class HinvSta(NSCL):

    TYPE = 'hinv_sta.d'

    def _initial_comments(self, comment):
        return []

    def __init__(self, file, prefer, sort, source):
        super().__init__(SNC_prefix_trailing_L, file, prefer, sort, source)


class PickSta(NSCL):

    TYPE = 'pick_sta.d'

    def __init__(self, file, prefer, sort, source):
        super().__init__(SCNL_after_2, file, prefer, sort, source)


class SlinkImports(NS):
    '''
    Old code, no longer used.  File contents are themselves patterns and must be applied to the NSCLS.
    See new code in slink.py
    '''

    TYPE = 'slink_imports.d'

    def __init__(self, file, prefer, sort, source):
        super().__init__(NS_after_Stream, file, prefer, sort, source)


class TrigSta(NSCL):

    TYPE = 'trigsta.scnl'

    def __init__(self, file, prefer, sort, source):
        super().__init__(SCNL_after_TrigStation, file, prefer, sort, source)


class Tbuf2Mseed(NSCL):

    TYPE = 'tbuf2mseed.d'

    def __init__(self, file, prefer, sort, source):
        super().__init__(SCNL_after_Send_scnl, file, prefer, sort, source)


class CarlSta(NSCL):

    TYPE = 'carlsta.scnl'

    def __init__(self, file, prefer, sort, source):
        super().__init__(SCNL_after_station, file, prefer, sort, source)

    def _write(self, file):
        for key, lines in self._sorted_items():
            net, sta, chan, loc = key.split('.')
            if not chan.endswith('Z'):
                log.warning(f'Some carlsta data contains non-Z channels')
            for line in lines:
                file.write(line)


class Fir(NSCL):

    TYPE = 'fir.scnl'

    def __init__(self, file, prefer, sort, source):
        super().__init__(SCNL_after_GetSCNL, file, prefer, sort, source)

    def _write(self, file):
        for key, lines in self._sorted_items():
            net, sta, chan, loc = key.split('.')
            if not chan.endswith('Z'):
                log.warning(f'Some fir data contains non-Z channels')
            for line in lines:
                file.write(line)

                
class BinderGrid(BINDER):
    ''' Handle Binder grid files.

    This parser just copies the content of all 
    binder_grid.d files into one. This doesn't make sense, 
    but it is used to avoid warnings, that no parser for binder_grid.d 
    files is available.
    The actual computation of the grid file for all merged stations
    is done in merge.binder.compute_binder_grid called by ewmerge.py.
    '''
    
    TYPE = 'binder_grid.d'

    def __init__(self, file, prefer, sort, source):
        super().__init__(binder_grid_coord_min_max, file, prefer, sort, source)

    def _write(self, file):
        for key, lines in self._sorted_items():
            for line in lines:
                file.write(line)


class WsvChanList(NSCL):

    TYPE = 'wsv_chan_list.d'

    def __init__(self, file, prefer, sort, source):
        super().__init__(SCNL_after_Tank, file, prefer, sort, source)


class StaList(NSCL):

    TYPE = 'stalist.txt'

    def __init__(self, file, prefer, sort, source):
        # always sorted
        super().__init__(SNCL, file, prefer, 'SNCL', source)


PARSERS[HinvSta.TYPE] = HinvSta
PARSERS['pick_FP_sta.d'] = PickSta
PARSERS[PickSta.TYPE] = PickSta
PARSERS[SlinkImports.TYPE] = SlinkImports
PARSERS[TrigSta.TYPE] = TrigSta
PARSERS[Tbuf2Mseed.TYPE] = Tbuf2Mseed
PARSERS[WsvChanList.TYPE] = WsvChanList
PARSERS[CarlSta.TYPE] = CarlSta
PARSERS[Fir.TYPE] = Fir
PARSERS[StaList.TYPE] = StaList
PARSERS[BinderGrid.TYPE] = BinderGrid




from argparse import ArgumentParser
from logging import getLogger
from os import mkdir
from os.path import exists
from shutil import copytree

from ewconfig.prune.response import delete_response
from ewconfig.prune.sacp import delete_sacp
from ewconfig.prune.xml import delete_xml
from .lib.args import add_version_args, ParagraphHelpFormatter
from .lib.file import assert_dir
from .lib.filter import DottedFilter
from .lib.log import add_log_args, make_log_from_args
from .lib.markdown import add_md_help_argument
from .merge.chan import ls_chan, PARSERS

log = getLogger(__name__)


def _prepare_dirs(old_dir, force, pruned, backup):
    if force:
        if (not pruned and not backup) or (pruned and backup):
            raise Exception('Provide exactly one of --pruned or --backup with --force')
        if pruned and exists(pruned):
            raise Exception(f'--pruned {pruned} already exists')
        if backup and backup != 'NONE' and exists(backup):
            raise Exception(f'--backup {backup} already exists')
    else:
        if pruned or backup:
            log.warning(f'No output until --force used (ignoring --pruned and --backup)')
    if force:
        if pruned:
            copytree(old_dir, pruned)
            return pruned
        else:  # backup
            if backup == 'NONE':
                log.warning(f'Mutating contents of {old_dir} with no backup')
            else:
                copytree(old_dir, backup)
                log.info(f'Backup in {backup}')
            return old_dir
    else:
        log.warning(f'Dry run - no output (use --force to change)')
        return old_dir


def main(old_dir, filter=DottedFilter('Pattern'), force=False, pruned=None, backup=None):
    dir = _prepare_dirs(old_dir, force, pruned, backup)
    log.debug(f'Working directory is {dir}')
    for file in ls_chan(dir):
        if file in PARSERS:
            Parser = PARSERS[file]
            log.debug(f'Parsing {file} with {Parser}')
            parser = Parser(file, None, False, file)
            parser.read(dir)
            parser.delete(filter, force=force)
            if force:
                parser.write(dir)
        else:
            log.warning(f'Ignoring {file}')
    delete_sacp(dir, filter, force=force)
    delete_response(dir, filter, force=force)
    delete_xml(dir, filter, force=force)
    return dir


def dotted(pattern):
    n = pattern.count('.')
    if n not in (1, 3):
        raise Exception(f'The pattern {pattern} does not contain 2 or 4 dotted names (N.S or N.S.C.L) so will not be used')
    return pattern


def main_args():
    try:
        parser = ArgumentParser(prog='ewprune', add_help=False,
                                formatter_class=ParagraphHelpFormatter,
                                description="""A tool remove specific sources from an EW configuration.
                                
To do this, sources are matched against patterns.  
Patterns are dotted components that can contain *, ? (wildcards), [...] and [!...] (alternatives).
Patterns only match NSCLs if the number of components matches.
So the pattern '*.*.HN[NE].*' will match CI.PASC.HNN.10 for example, but will not match CI.PASC 
(without channel or location).
Multiple patterns can be specified and sources are only deleted if they match a pattern in `--delete` 
and do not match a pattern in `--keep`.

This may seem cumbersome, but seems to be the only approach that gives complete control.

By default the sources to be deleted are displayed (only).
To actually force deletion you must add `--force`.

If --pruned is given, the new, pruned config is placed in that directory.

If --pruned is not given, the pruned config replaces the contents of the --dir directory.  
In this case --backup (which will contain the original contents of --dir) must be given.  
If you really don't want a backup use "--backup NONE".

Examples

To delete all mention of the CI.PASC station (both N.S and N.S.C.L entries in various files):

ewprune DIR --delete CI.PASC CI.PASC.*.* --backup ...

To remove all HHZ channels 

ewprune DIR --delete *.*.HHZ.* --backup ... 

Note that in the example above N.S entries will remain, so if station NET.XXX had only HHZ then that must 
also be explicitly deleted:

ewprune DIR --delete *.*.HHZ.* NET.XXX --backup ... 

Alternatively, maybe we know that only NET.YYY had alternative channels, 
so we can delete all N.S entries except that:

ewprune DIR --delete *.*.HHZ.* *.* --keep NET.YYY --backup ... 
""")
        parser.add_argument('dir', metavar='DIR', type=assert_dir,
                            help='Directory containing the configuration')
        parser.add_argument('--delete', metavar='PATTERN', type=dotted, nargs='+', help='Patterns to delete')
        parser.add_argument('--keep', metavar='PATTERN', type=dotted, nargs='+', help='Patterns to keep')
        parser.add_argument('--force', action='store_true', help='Force deletion')
        parser.add_argument('--pruned', help='Destination directory for pruned config.')
        parser.add_argument('--backup', help='Backup directory for original config (if no --pruned; can be NONE).')
        parser.add_argument('-h', action='help', help='Show this help message and exit.')
        add_version_args(parser)
        add_md_help_argument(parser)
        add_log_args(parser)
        args = parser.parse_args()
        make_log_from_args(parser.prog, args)
        main(args.dir, filter=DottedFilter('Pattern', includes=args.delete, excludes=args.keep),
             force=args.force, pruned=args.pruned, backup=args.backup)
    except Exception as e:
        log.critical(e)
        exit(1)


if __name__ == '__main__':
    main_args()



from glob import glob
from logging import getLogger
from os import mkdir, scandir
from os.path import join, exists, basename, splitext
from shutil import copyfile

from obspy import read_inventory

from ewconfig.merge.prefer import Prefer

log = getLogger(__name__)


def merge_xml(old, new, merged, prefer, subdir='stationxml'):
    src_old = join(old, subdir) if old else None
    src_new = join(new, subdir) if new else None
    dest = join(merged, subdir)
    if prefer == Prefer.OLD:
        _merge_xml_recursive(src_old, src_new, dest, False)
    elif prefer == Prefer.NEW:
        _merge_xml_recursive(src_new, src_old, dest, False)
    else:
        _merge_xml_recursive(src_new, src_old, dest, True)


def _merge_xml_recursive(over, base, dest, unique):

    def make_dest_on_demand():
        if not exists(dest):
            mkdir(dest)

    def xml_files_in(dir):
        yield from (basename(file) for file in glob(join(dir, '*.xml')))

    def subdirs_in(dir):
        yield from (basename(dir.path) for dir in scandir(dir) if dir.is_dir())

    known = set()
    if base and exists(base):
        make_dest_on_demand()
        for file in xml_files_in(base):
            known.add(file)
            log.debug(f'Copying {file} from {base} to {dest}')
            copyfile(join(base, file), join(dest, file))
    if over and exists(over):
        make_dest_on_demand()
        for file in xml_files_in(over):
            if unique and file in known:
                if read_inventory(join(base, file)) == read_inventory(join(over, file)):
                    log.debug(f'{file} in both {over} and {base} but identical')
                else:
                    raise Exception(f'{file} in {over} and {base} differ (use --prefer-new or --prefer-old)')
            else:
                log.debug(f'Copying {file} from {over} to {dest}')
                copyfile(join(over, file), join(dest, file))
    known = set()
    if base and exists(base):
        make_dest_on_demand()
        for dir in subdirs_in(base):
            known.add(dir)
            _merge_xml_recursive(join(over, dir) if over else None, join(base, dir), join(dest, dir), unique)
    if over and exists(over):
        make_dest_on_demand()
        for dir in subdirs_in(over):
            if dir not in known:
                _merge_xml_recursive(join(over, dir), join(base, dir) if base else None, join(dest, dir), unique)



from argparse import ArgumentParser
from logging import getLogger
from os import getcwd, mkdir, unlink
from os.path import sep, join, exists

from obspy import read_inventory
from obspy.io.stationxml.core import validate_stationxml

from .lib.args import ParagraphHelpFormatter, add_version_args
from .lib.date import parse_date, between
from .lib.file import assert_dir
from .lib.filter import Filter
from .lib.log import add_log_args, make_log_from_args
from .lib.markdown import add_md_help_argument
from .lib.write import write_all, add_write_args, DEFAULT_PICK_FP_STA
from .pz.xml import XmlPolesZeros, Response, UnsupportedResponseException

'''
The equivalent of sacpz2ew, but takes input from station.xml files. 
'''

log = getLogger(__name__)


def _validate(file):
    ok, errors = validate_stationxml(file)
    if not ok:
        log.info('The following warnings are from XML validation:')
        for error in errors:
            log.warning(error)
        # paul's Ince_v1 file triggers warnings
        log.info('End of XML validation warnings')
        log.info('Will continue despite warnings above')


def _nscl(network, station, channel):
    return f'{network.code}.{station.code}.{channel.code}.' \
           f'{channel.location_code if channel.location_code else "--"}'


def _filter(metadata, sta_filter, chan_filter, date=None):
    # warning - this actually mutates the metadata
    for network in metadata:
        for station in list(network):
            sta_used = False
            if sta_filter.test(station, key=lambda sta: sta.code):
                for channel in list(station):
                    chan_used = False
                    nscl = _nscl(network, station, channel)
                    if chan_filter.test(channel, key=lambda chan: chan.code):
                        if between(channel.start_date, date, channel.end_date):
                            log.debug(f'Keeping {nscl} ({date if date else "now"} included)')
                            chan_used = True
                            sta_used = True
                    if not chan_used:
                        log.warning(f'Deleting {nscl}: {date} not in {channel.start_date} - {channel.end_date}')
                        station.channels.remove(channel)
            if not sta_used:
                log.info(f'Deleting station {station} because it was either filtered by command line args, '
                         f'or it has no channels')
                network.stations.remove(station)
    return metadata


def _parse(metadata, lo_precision=False, default_response=None):
    for network in metadata.networks:
        for station in network.stations:
            for channel in station.channels:
                nscl = _nscl(network, station, channel)
                try:
                    log.debug(f'Processing {nscl}')
                    yield XmlPolesZeros(network, station, channel,
                                        lo_precision=lo_precision, default_response=default_response)
                except UnsupportedResponseException as e:
                    log.warning(f'Skipping {nscl}: {e}')


def main_multiple(files, dirs=None, xml_outs=None, drop_comment=False, m_to_nm=1, geophone=False,
                  lo_precision=False, default_response=None, date=None, pick_fp_sta=DEFAULT_PICK_FP_STA,
                  sta_filter=Filter('Station'), chan_filter=Filter('Channel')):
    if len(files) == 1 and not dirs:
        dirs = [None]
    if len(dirs) != len(files):
        raise Exception('The number of output directories must match the number of XML files.')
    if not xml_outs:
        # None is set to a value later (based on filtered stations)
        xml_outs = [None] * len(files)
    if len(xml_outs) != len(files):
        raise Exception('If output files are given (--xml-out) the number of files must match the number of '
                        'input XML files.')
    if any(sep in xml_out for xml_out in xml_outs if xml_out):
        raise Exception('Output XML (--xml-out) files must not contain slashes (they go in the stationxml dir)')
    for file, dir, xml_out in zip(files, dirs, xml_outs):
        log.info(f'Expanding {file} into {dir}')
        main_single(file, dir=dir, xml_out=xml_out, drop_comment=drop_comment, m_to_nm=m_to_nm, geophone=geophone,
                    lo_precision=lo_precision, default_response=default_response, date=date, pick_fp_sta=pick_fp_sta,
                    sta_filter=sta_filter, chan_filter=chan_filter)


def _first_sta(net):
    stas = sorted((sta for sta in net), key=lambda sta: sta.code)
    return net.code + '-' + stas[0].code


def _default_xml_name(metadata):
    nets = sorted((net for net in metadata), key=lambda net: net.code)
    name = _first_sta(nets[0])
    if len(nets) > 1:
        name += '-' + _first_sta(nets[-1])
    return name + '.xml'


def __fix_output_dir(dir, subdir):
    dir = join(dir, subdir)
    if not exists(dir):
        mkdir(dir)
    return dir


def _copy_xml_monolith(metadata, dir, xml_out, subdir='stationxml'):
    if not xml_out: xml_out = _default_xml_name(metadata)
    dir = __fix_output_dir(dir, subdir)
    xml_out = join(dir, xml_out)
    log.info(f'Writing filtered XML to {xml_out}')
    metadata.write(xml_out, format='STATIONXML')


def _copy_xml_by_net_sta(metadata, dir, subdir='stationxml'):
    dir = __fix_output_dir(dir, subdir)
    for net in metadata:
        net_dir = __fix_output_dir(dir, net.code)
        if not exists(net_dir):
            mkdir(net_dir)
        net_metadata = metadata
        for nm_net in list(net_metadata):
            if nm_net.code != net.code:
                net_metadata = net_metadata.remove(network=nm_net.code)
        for sta in net:
            sta_path = join(net_dir, sta.code + '.xml')
            if exists(sta_path):
                log.warning(f'Overwriting {sta_path}')
                unlink(sta_path)
            sta_metadata = net_metadata
            for sm_sta in list(sta_metadata[0]):
                if sm_sta.code != sta.code:
                    sta_metadata = sta_metadata.remove(network=net.code, station=sm_sta.code)
            log.info(f'Writing XML to {sta_path}')
            sta_metadata.write(sta_path, format='STATIONXML')


def main_single(file, dir=None, xml_out=None, drop_comment=False, m_to_nm=1, geophone=False,
                lo_precision=False, default_response=None, date=None, pick_fp_sta=DEFAULT_PICK_FP_STA,
                sta_filter=Filter('Station'), chan_filter=Filter('Channel')):
    log.debug(f'Calling stationxml2ew single for file={file} dir={dir} drop_comment={drop_comment} '
              f'm_to_nm={m_to_nm} geophone={geophone} lo_precision={lo_precision} '
              f'default_response={default_response}sta_filter={sta_filter} chan_filter={chan_filter} '
              f'date={date}')
    _validate(file)
    dir = assert_dir(dir or getcwd(), 'Output dir (--dir)', create=True)
    metadata = read_inventory(file)
    metadata = _filter(metadata, sta_filter, chan_filter, date=date)
    nscls = list(_parse(metadata, lo_precision=lo_precision, default_response=default_response))
    write_all(dir, nscls, drop_comment=drop_comment, m_to_nm=m_to_nm, geophone=geophone, pick_fp_sta=pick_fp_sta)
    if xml_out:
        _copy_xml_monolith(metadata, dir, xml_out)
    else:
        _copy_xml_by_net_sta(metadata, dir)

    
def add_stationxml2ew_args(parser):
    # separated here so that it can also be used by ewmerge
    parser.add_argument('--include-chan', metavar='PATTERN', nargs='*',
                        help='Pattern to match included channel names (default all).')
    parser.add_argument('--exclude-chan', metavar='PATTERN', nargs='*',
                        help='Pattern to match excluded channel names.')
    parser.add_argument('--include-sta', metavar='PATTERN', nargs='*',
                        help='Pattern to match included station names (default all).')
    parser.add_argument('--exclude-sta', metavar='PATTERN', nargs='*',
                        help='Pattern to match excluded station names.')
    parser.add_argument('--date', type=parse_date, help='Use channels valid on this date (default now).')
    parser.add_argument('--pick-fp-sta', metavar='STRING',
                        default=DEFAULT_PICK_FP_STA,
                        help=f'Extra values for pick_FP_sta.d (default "{DEFAULT_PICK_FP_STA}").')
    defaults = parser.add_mutually_exclusive_group()
    defaults.add_argument('--displacement', dest='default_response', action='store_const',
                          const=Response.DISPLACEMENT, help='Assume unrecognized responses are displacement.')
    defaults.add_argument('--velocity', dest='default_response', action='store_const',
                          const=Response.VELOCITY, help='Assume unrecognized responses are velocity.')
    defaults.add_argument('--acceleration', dest='default_response', action='store_const',
                          const=Response.ACCELERATION, help='Assume unrecognized responses are acceleration.')
    parser.set_defaults(default_response=None)
    add_write_args(parser, with_drop_comment=False)


def main_args():
    args = None  # used in cleanup
    try:
        parser = ArgumentParser(prog='stationxml2ew',
                                formatter_class=ParagraphHelpFormatter,
                                description='''A scanner program to 
convert a station.xml file to configurations suitable for use with localmag, hypoinverse, 
pick_ew, pick_FP, slink2ew, carlstatrig and wave_serverV station lists. It is presumed 
that the input file is complete, with lat/long/elevation for the SCNL.  Output sac files 
will be in meters unless the --nano parameter is set to convert to nanometers. 

Stations and channels can be selected using --include-chan (-sta) and --exclude-chan (-sta).
These take patterns that can match any string (*), any character (?), and any character from 
a sequence ([abc]).  If --include-chan (-sta) is not given it is assumed to be "*".  Channels,
for example, are included if they match --include-chan and do not match --exclude-chan.

Specifying --date selects channels active at that time.  By default this is the current time.

Responses types are recognised via the response units.  Unrecognised units can be given a default
using --displacement|--velocity|--acceleration.  If no default is given they are skipped.  Pressure
sensitive stations (units Pa) are always skipped.    

Examples

To generate a config in the current directory from an XML file:

stationxml2ew station.xml

To generate a config in a different directory:

stationxml2ew station.xml --dir output/dir

To process multiple XML files, each with a different output directory:

stationxml2ew station1.xml station2.xml --dir station1 station2

To convert multiple XML files into a single directory see ewmerge (which can
call this program and then do the merge to a single directory).
''')
        parser.add_argument('file', nargs='+', default=[], metavar='FILE',
                            help='Station XML file(s)')
        parser.add_argument('--dir', nargs='*', default=[], metavar='DIR',
                            help='Output directory(s) (default CWD for a single input file)')
        parser.add_argument('--xml-out', nargs='+', metavar='FILE', help='Filtered station.xml output')
        add_stationxml2ew_args(parser)
        add_version_args(parser)
        add_md_help_argument(parser)
        add_log_args(parser)
        args = parser.parse_args()
        make_log_from_args(parser.prog, args)
        main_multiple(args.file, dirs=args.dir, xml_outs=args.xml_out,
                      drop_comment=args.drop_comment, m_to_nm=args.nano, geophone=args.geophone,
                      default_response=args.default_response, date=args.date, pick_fp_sta=args.pick_fp_sta,
                      sta_filter=Filter('Station', includes=args.include_sta, excludes=args.exclude_sta),
                      chan_filter=Filter('Channel', includes=args.include_chan, excludes=args.exclude_chan))
    except Exception as e:
        log.critical(e)
        if args and args.verbosity == 5:
            log.exception(e)
        exit(1)


if __name__ == '__main__':
    main_args()



from glob import glob
from logging import getLogger
import os

import numpy as np
from obspy import read_inventory
import obspy.geodetics.base as geodetics

from ewconfig.pz.xml import XmlPolesZeros, UnsupportedResponseException
# Import the module to avoid circular import errors.
import ewconfig.lib.write as write

log = getLogger(__name__)


def compute_grid_limits(lon, lat, border_win_ratio = 0.25,
                        min_border = 0.1, max_border = 10,
                        round_dec = 2):
    ''' Compute the binder grid limits from station coordinates.
    
    Parameters
    ---------.
    lon: :obj:`list` of float
        The station longitude values.
        
    lat: :obj:`list` of float
        The station latitude values.
        
    border_win_ratio: float
        The ratio of the total extent used to add a border
        to the grid (0 -1)
        
    min_border: float
        The minimum border width [degree].
        
    max_border: float
        The maximum border width [degree].

    round_dec: None or int
        Round to these number of decimals.
        If None, the values are not rounded.
        
    Returns
    -------
    :obj:`list` of float (2x2)
        The grid limits.
        [[lon_min, lon_max],
         [lat_min, lat_max]
    '''
    # Get the coordinate limits.
    min_lat = min(lat)
    max_lat = max(lat)
    min_lon = min(lon)
    max_lon = max(lon)

    # Compute the border width.
    border_lat = (max_lat - min_lat) * border_win_ratio
    border_lon = (max_lon - min_lon) * border_win_ratio

    if border_lat < min_border:
        border_lat = min_border

    if border_lat > max_border:
        border_lat = max_border

    if border_lon < min_border:
        border_lon = min_border

    if border_lon > max_border:
        border_lon = max_border

    # Add the border.
    grid_limits = [min_lat - border_lat,
                   max_lat + border_lat,
                   min_lon - border_lon,
                   max_lon + border_lon]

    # Check coordinate limits.
    if (grid_limits[0] < -90):
        grid_limits[0] = -90

    if(grid_limits[1] > 90):
        grid_limits[1] = 90

    if(grid_limits[2] < -180):
        grid_limits[2] = -180

    if(grid_limits[3] > 180):
        grid_limits[3] = 180

    if round_dec is not None:
        grid_limits = [round(x, round_dec) for x in grid_limits]
 
    return grid_limits


def compute_grid_parameters(lon, lat):
    ''' Compute the binder grid parameters.

    Parameters
    ---------.
    lon: :obj:`list` of float
        The station longitude values.
  
    lat: :obj:`list` of float
        The station latitude values.

    Returns
    -------
    :obj:`dictionary`
        The computed grid parameters with the parameter names as keys.
    '''
    ret = {}
    ret['dspace'] = 4.0
    ret['rstack'] = 100.0
    ret['tstack'] = 0.5
    stat_coords = np.array(list(zip(lon, lat)))
    stat_coords = np.unique(stat_coords, axis = 0)

    n_stations = len(stat_coords)

    # Return the default values if less than 2 stations
    # are passed.
    if n_stations <= 1:
        return ret

    dist = []
    for k in range(n_stations - 1):
        cur_main = stat_coords[k, :]
        for m in range(k + 1, n_stations):
            cur_ref = stat_coords[m, :]
            cur_dist = geodetics.gps2dist_azimuth(cur_main[1], cur_main[0],
                                                  cur_ref[1], cur_ref[0])
            dist.append(cur_dist[0])
    dist = np.array(dist)

    #log.info('#### Computing grid parameters.')
    #log.info('#### dist: {}.'.format(dist))
    #log.info('#### n_stations: {}.'.format(n_stations))
    #log.info('#### stat_coords: {}.'.format(stat_coords))
    
    # Compute the binder parameters.
    # dspace
    ratio = 0.25
    min_dist = np.min(dist)
    dspace = np.floor((min_dist * ratio) / 1000) * 1000

    # rstack
    factor = 0.75
    med_dist = np.median(dist)
    rstack = (med_dist - (med_dist % dspace)) * factor

    # tstack
    # As a first approximation, tstack should be at least the number
    # of seconds it takes for a P-wave to cross a grid cell of dspace km.
    # Compute the traveltime using a constant vp velocity.
    ncells = 2
    vp = 4000
    tstack = (dspace * ncells) / vp

    ret = {}
    ret['dspace'] = dspace / 1000
    ret['rstack'] = rstack / 1000
    ret['tstack'] = tstack

    return ret


def compute_binder_grid(old, subdir='stationxml'):
    ''' Compute the binder grid using stationXML files.
    '''
    def xml_files_in(dir):
        search_path = os.path.join(dir, '**', '*.xml')
        yield from (file for file in glob(search_path, recursive = True))
        
    xml_path = os.path.join(old, subdir)

    # Read all available station XML files of the merged directory.
    inv = None
    for file in xml_files_in(xml_path):
        cur_inv = read_inventory(file)
        if inv is None:
            inv = cur_inv
        else:
            inv.extend(cur_inv)

    if not inv:
        log.warning(f'No station XML files in {xml_path}')
    else:
        # Compute the grid limits.
        sncls = []
        for cur_net in inv:
            for cur_stat in cur_net:
                for cur_chan in cur_stat:
                    try:
                        cur_pz = XmlPolesZeros(cur_net, cur_stat, cur_chan,
                                               lo_precision=False,
                                               default_response=None)
                        sncls.append(cur_pz)
                    except UnsupportedResponseException:
                        log.warning(f'Skipping {cur_stat}.{cur_net}.{cur_chan} (pressure sensitive detector)')

        eqk_dir = os.path.join(old, 'eqk')
        write.write_binder_grid(eqk_dir, sncls)
    

    



from abc import ABC
from logging import getLogger
from os import makedirs
from os.path import join, exists

log = getLogger(__name__)


class BaseParser(ABC):

    def __init__(self, file, source):
        self._file = file
        self._source = source
        self.__deletion = False

    def read(self, dir, comment=None):
        path = join(dir, 'chan', self._file)
        if exists(path):
            log.debug(f'Reading {path}')
            with open(path, 'r') as src:
                self._read(src, comment=comment)
        else:
            log.warning(f'Could not find {path}')

    def _read(self, file, comment=None):
        raise NotImplementedError('_read')

    def add(self, new):
        raise NotImplementedError('add')

    def write(self, merged):
        dir = join(merged, 'chan')
        if not exists(dir):
            makedirs(dir)
        path = join(dir, self._file)
        log.info(f'Writing {path}')
        with open(path, 'w') as file:
            self._write(file)

    def _write(self, file):
        raise NotImplementedError('_write')

    def _print_deletion(self, lines):
        # presumably there was a reason this prints rather than logs
        # i guess so that it uses stdout rather than stderr?
        if not self.__deletion:
            print(f'DELETIONS in {self._source}:')
            self.__deletion = True
        for line in lines:
            print('  ', line, end='' if line.endswith('\n') else '\n')

    def _end_deletions(self):
        if not self.__deletion:
            print(f'UNCHANGED {self._source}')
